/* DIY 3D NAVIGATOR FOR BLENDER V02
by meisterodin (March 2024)

Sketch for 3d navigator using a MPU6050 to detect angle and generate mouse movement and
a rotary encoder for scrolling; as well as 8 keys to generate keyboard commands and
an optional array of neopixels to display the Blender color palette (blue, orange 
and white, randomly selected at startup).

It is designed to be used with a Teensy 2 (with teensyduino) using the Teensy keyboard
library and its key codes (https://www.pjrc.com/teensy/td_keyboard.html). It can be
adapted to work with Arduino boards with built-in HID support, such as the Arduino Leonardo 
and the Arduino MICRO using a ATMega32u4 MCU, but I suppose it must use the Arduino key codes
(they're slightly different).

The keys are mapped using the OneButton library. It allows to add functions for single,
double, long press and long press release.

Get instructions and files at https://www.instructables.com/DIY-3D-Navigator-for-Blender/

Get stl and blend files at https://www.thingiverse.com/thing:6555490/files.


*** NOTE *** 
It is important to set USB type to Keyboard + Mouse + Joystick in "Tools" >> "USB Type" in
the arduino software, as well as the desired keyboard layout in "Tools" >> "Kayboard layout"

-------------------------------------------------------------------------------------------

Sources:

The MPU6050 section is based on the excellent 3D Space Rat by KikiHobbyRepair
https://github.com/KikiHobbyRepair/SpaceRat, using the MPU6050_light library
by rfetick https://github.com/rfetick/MPU6050_light

The buttons interface was adapted from the OneButton library by Matthias Hertel,
https://github.com/mathertel/OneButton 

The RotaryEncoder library, also by Matthias Hertel,
https://github.com/mathertel/RotaryEncoder 

The neopixel section was adapted from Adafruit's Neopixel sketch,
https://github.com/adafruit/Adafruit_NeoPixel 

-------------------------------------------------------------------------------------------
*/

// LIBRARIES AND VARIABLES
#include <Wire.h>
#include <MPU6050_light.h>
#include <Mouse.h>
#include <Keyboard.h>
#include "RotaryEncoder.h"
#include "OneButton.h"     
#include <Adafruit_NeoPixel.h>

// ROTARY ENCODER PINS
RotaryEncoder encoder(8, 9);

//MPU6050 CONFIG
MPU6050 mpu(Wire);
unsigned long timer = 0;
int Xang, Yang, Zang;

//SCROLL WHEEL POSITION
static int pos = 0;
int newPos;
int scrollPos = 0;
bool invertZoom = false; //Invert scrolling direction


String layout = "QWERTY"; // DEFAULT KEYBOARD LAYOUT SELECTION: "ENGLISH", "FRENCH" or "SPANISH"

// KEYBOARD VARIABLES (BLENDER)
int m;
int a;
int q;
int g;
int r;
int s;
int c;
int x;
int y;
int z;
int b;

//KEYS AND BUTTONS PINS

//     ----------- ----------- ----------- -----------
//    |           |           |           |           |
//    |   KEY1    |   KEY2    |    KEY3   |   KEY4    |
//    |  pin 17   |  pin 16   |   pin 15  |  pin 14   |
//     ----------- ----------- ----------- -----------
//    |           |           |           |           |
//    |   KEY5    |   KEY6    |    KEY7   |   KEY8    |
//    |  pin 18   |  pin 19   |   pin 20  |   pin 21  |
//     ----------- ----------- ----------- -----------
//         *  *                                   *  *   
//      *        *                             *        * 
//     * ButtonL  *                           *  ButtonR * 
//     *  pin 3   *                           *  pin 0   * 
//      *        *                             *        *
//         *  *                                   *  *

OneButton key1(17, true); 
OneButton key2(16, true);
OneButton key3(15, true);
OneButton key4(14, true);
OneButton key5(18, true);
OneButton key6(19, true);
OneButton key7(20, true);
OneButton key8(21, true);
OneButton buttonR(0, true);
OneButton buttonL(3, true);

int rotaryEncoderButton = 7; // Rotary encoder button pin
int encoderButtonState = 0; // Rotary encoder button state

//NEOPIXEL VARIABLES
int color; // Color variable (Blender palette): 0 = blue; 1 = orange; 2 = white.

// Pins used to drive the pixels
int boardPin =  10;
int ringPin =  4;

// Number of pixels
int numBoardPixels = 5; // Board pixels
int numRingPixels = 3; // Ring pixels

// NeoPixel color format & data rate
int pixelFormat = NEO_GRB + NEO_KHZ800;

// Neopixel pointer object, allocated in SETUP
Adafruit_NeoPixel *boardPixels ;
Adafruit_NeoPixel *ringPixels ;

//#define DELAYVAL 0 // Time (in milliseconds) to pause between pixels

// END OF LIBRARIES AND VARIABLES

void setup() { // SETUP

  //NEOPIXEL SETUP
  randomSeed(analogRead(A11)); // Random seed generator assigned to analog pin A11 
  color = random(3); //Choose randomly from blue, orange and white

  // Create a new NeoPixel objects
  boardPixels = new Adafruit_NeoPixel(numBoardPixels, boardPin, pixelFormat);
  ringPixels = new Adafruit_NeoPixel(numRingPixels, ringPin, pixelFormat);

  // INITIALIZE NeoPixel objets
  boardPixels->begin(); 
  ringPixels->begin();
  
  pinMode(rotaryEncoderButton, INPUT_PULLUP); //Assign rotary encoder button pin
  Wire.begin(); //Initialize I2C
  mpu.begin(); //Initialize MPU
  //delay(1000);
  startUpColor(); // Starutup color animation. Cycles blue, orange, green and off for 1 second.
  mpu.calcOffsets(); // Calculate MPU offsets
  Mouse.begin(); // Initialize mouse interface
  Keyboard.begin(); // Initialize keyboard interface


  
  //########### KEY & BUTTON SETUP ############
  // link the key1 functions.
  key1.attachClick(singleClick_1);
  key1.attachDoubleClick(doubleClick_1);
  key1.attachLongPressStop(longPressStop_1);
  key1.attachDuringLongPress(longPress_1);

  // link the key2 functions.
  key2.attachClick(singleClick_2);
  key2.attachDoubleClick(doubleClick_2);
  key2.attachLongPressStop(longPressStop_2);
  key2.attachDuringLongPress(longPress_2);

  // link the key3 functions.
  key3.attachClick(singleClick_3);
  key3.attachDoubleClick(doubleClick_3);
  key3.attachLongPressStop(longPressStop_3);
  key3.attachDuringLongPress(longPress_3);

  // link the key4 functions.
  key4.attachClick(singleClick_4);
  key4.attachDoubleClick(doubleClick_4);
  key4.attachLongPressStop(longPressStop_4);
  key4.attachDuringLongPress(longPress_4);

    // link the key5 functions.
  key5.attachClick(singleClick_5);
  key5.attachDoubleClick(doubleClick_5);
  key5.attachLongPressStop(longPressStop_5);
  key5.attachDuringLongPress(longPress_5);

    // link the key 6 functions.
  key6.attachClick(singleClick_6);
  key6.attachDoubleClick(doubleClick_6);
  key6.attachLongPressStop(longPressStop_6);
  key6.attachDuringLongPress(longPress_6);

      // link the key 7 functions.
  key7.attachClick(singleClick_7);
  key7.attachDoubleClick(doubleClick_7);
  key7.attachLongPressStop(longPressStop_7);
  key7.attachDuringLongPress(longPress_7);

      // link the key 8 functions.
  key8.attachClick(singleClick_8);
  key8.attachDoubleClick(doubleClick_8);
  key8.attachLongPressStop(longPressStop_8);
  key8.attachDuringLongPress(longPress_8);

     // link the button R functions.
  buttonR.attachClick(singleClick_buttonR);
  buttonR.attachDoubleClick(doubleClick_buttonR);
  buttonR.attachLongPressStop(longPressStop_buttonR);
  buttonR.attachDuringLongPress(longPress_buttonR);

      // link the button L functions.
  buttonL.attachClick(singleClick_buttonL);
  buttonL.attachDoubleClick(doubleClick_buttonL);
  buttonL.attachLongPressStop(longPressStop_buttonL);
  buttonL.attachDuringLongPress(longPress_buttonL);


} // END OF SETUP


void loop() {

  encoderButtonState = digitalRead(rotaryEncoderButton); // Read rotary encoder button state
 
  pixelColor(); //Display neopixel colors

  // Define layout language
  layout = "QWERTY"; 

  if (layout == "QWERTY"){
    a = 97;
    q = 113;
    m = 109;
  }

  else if (layout == "AZERTY"){
    a = 113;
    q = 97;
    m = 61491;
  }
  
  mpu.update(); // Get data from MPU6050

  // GETTING KEYS, BUTTON AND ROTARY ENCODER EVENTS
  key1.tick();
  key2.tick();
  key3.tick();
  key4.tick();
  key5.tick();
  key6.tick();
  key7.tick();
  key8.tick();
  buttonR.tick();
  buttonL.tick();
  encoder.tick();

  newPos = encoder.getPosition(); // New encoder position
        
  if ((millis() - timer) > 25) { 

    //Reading the MPU data
    Xang=mpu.getAngleX();
    Yang=mpu.getAngleY();
    Zang=mpu.getAngleZ();

    //Logic for the Mouse and Keyboard movements
  if (encoderButtonState == 1)  {
      if  (Yang > 6) {  
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(map(Yang, 6, 10, 1, 10), 0 , 0);
      }

      else if (Yang < -6) {
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(map(Yang, -10, -6, -10, -1), 0 , 0); 
      }
      
    
      else if  (Xang > 6) {  
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(0, map(Xang, 6, 10, 1, 10) , 0, 0); 
      }

      else if (Xang < -6) {
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(0, map(Xang, -10, -6, -10, -1) , 0, 0); 
      }

      // Scrolling wheel action
      else if ( newPos > pos) {
        pos = newPos;
        if (invertZoom == false){
          scrollPos = -1;
          Mouse.move(0, 0, scrollPos);
        }

        else if (invertZoom == true){
          scrollPos = 1;
          Mouse.move(0, 0, scrollPos);
        }
      }

      else if ( pos > newPos){ 
        pos = newPos; 
        if (invertZoom == false){
          scrollPos = 1;
          Mouse.move(0, 0, scrollPos);
        }
        else if (invertZoom == true){
          scrollPos = -1;
          Mouse.move(0, 0, scrollPos);
        }        
      }

      else  {   
          Mouse.release(MOUSE_MIDDLE);

      }

  }

  else if (encoderButtonState == 0)   {
      if  (Yang > 6) {
        Keyboard.press(KEY_RIGHT_SHIFT);  
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(map(Yang, 6, 10, 1, 10), 0 , 0); 
      }

      else if (Yang < -6) {
        Keyboard.press(KEY_RIGHT_SHIFT);
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(map(Yang, -10, -6, -10, -1), 0 , 0); 
      }
      
    
      else if  (Xang > 6) {  
        Keyboard.press(KEY_RIGHT_SHIFT);
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(0, map(Xang, 6, 10, 1, 10) , 0, 0); 
      }

      else if (Xang < -6) {
        Keyboard.press(KEY_RIGHT_SHIFT);
        Mouse.press(MOUSE_MIDDLE);
        Mouse.move(0, map(Xang, -10, -6, -10, -1) , 0, 0); 
      }


      else  {   
          Mouse.release(MOUSE_MIDDLE);
          Keyboard.release(KEY_RIGHT_SHIFT);
      }
  }
    
    timer = millis();
  }
            
} // END OF MAIN VOID LOOP


// ############ FUNCTIONS ###############

// KEY1 FUNCTIONS
void singleClick_1() {
  Keyboard.press(KEYPAD_7);
  Keyboard.releaseAll();
} // End of singleClick_1

void doubleClick_1() {

  Keyboard.press(KEY_RIGHT_CTRL);
  Keyboard.press(KEYPAD_7);
  Keyboard.releaseAll();

} // End of doubleClick_1

void longPress_1() {
  Keyboard.press(KEYPAD_5);
} // End of longPress_1

void longPressStop_1() {
  Keyboard.releaseAll();
}

// KEY2 FUNCTIONS
void singleClick_2() {
  Keyboard.press(KEYPAD_1);
  Keyboard.releaseAll();
} // End of singleClick_2

void doubleClick_2() {
  Keyboard.press(KEY_RIGHT_CTRL);
  Keyboard.press(KEYPAD_1);
  Keyboard.releaseAll();
} // End of doubleClick_2

void longPress_2() {
  Keyboard.press(KEYPAD_PERIOD);  

} // End of longPress_2

void longPressStop_2() {
  Keyboard.releaseAll();
} // End of longPressStop_2 {


// KEY3 FUNCTIONS
void singleClick_3() {
  
  Keyboard.press(KEYPAD_3);
  Keyboard.releaseAll();

} // End of singleClick_3

void doubleClick_3() {

  Keyboard.press(KEY_RIGHT_CTRL);
  Keyboard.press(KEYPAD_3);
  Keyboard.releaseAll();

} // End of doubleClick_3

void longPress_3() {
  Keyboard.press(KEYPAD_SLASH);
} // End of longPress_3

void longPressStop_3() {
  Keyboard.releaseAll();
} // End of longPressStop_3

// KEY4 FUNCTIONS
void singleClick_4() {
  Keyboard.press(KEY_TAB);
  Keyboard.releaseAll();
} // End of singleClick_4

void doubleClick_4() { 
  Keyboard.press(q);
  Keyboard.releaseAll();
} // End of doubleClick_4

void longPress_4() {
  
} // End of longPress_4

void longPressStop_4() {
  Keyboard.releaseAll();
} // End of longPressStop_4

// KEY5 FUNCTIONS
void singleClick_5() {
  Keyboard.press('x');
  Keyboard.releaseAll();
} // End of singleClick_5

void doubleClick_5() {
  Keyboard.press('g');
  Keyboard.releaseAll();
} // End of doubleClick_5

void longPress_5() {
  Keyboard.press(KEY_RIGHT_SHIFT);
  Keyboard.press('x');
} // End of longPress_5

void longPressStop_5() {
  Keyboard.releaseAll();
  
} // End of longPressStop_5

// KEY6 FUNCTIONS
void singleClick_6() {
  Keyboard.press('y');
  Keyboard.releaseAll();
} // singleClick_6

void doubleClick_6() {
  Keyboard.press('s');
  Keyboard.releaseAll();
} // End of doubledoubleClick_6click6

void longPress_6() {
  Keyboard.press(KEY_RIGHT_SHIFT);
  Keyboard.press('y');
} // End of longPress_6

void longPressStop_6() {
  Keyboard.releaseAll();
} // End of longPressStop_6


// KEY7 FUNCTIONS
void singleClick_7() {
  Keyboard.press('z');
  Keyboard.releaseAll();
} // singleClick_7

void doubleClick_7() {
  Keyboard.press('r');
  Keyboard.releaseAll();
} // End of doubleClick_7


void longPress_7() {
  Keyboard.press(KEY_RIGHT_SHIFT);
  Keyboard.press('z');
} // End of longPress_7

void longPressStop_7() {
  Keyboard.releaseAll();
} // End of longPressStop_7


// KEY8 FUNCTIONS
void singleClick_8() {
  Keyboard.press(a);
  Keyboard.releaseAll();
} // singleClick_8

void doubleClick_8() {
  Keyboard.press(KEY_LEFT_ALT);
  Keyboard.press(a);
  Keyboard.releaseAll();
} // End of doubleClick_8


void longPress_8() {
  Keyboard.press('b');
} // End of longPress_8

void longPressStop_8() {
  Keyboard.releaseAll();
} // End of longPressStop_8


// BUTTON R FUNCTIONS
void singleClick_buttonR() {
  Keyboard.press('c');
} // End of singleClick_buttonR

void doubleClick_buttonR() {
  Keyboard.press('c');
  Keyboard.releaseAll();
} // End of doubleClick_buttonR

void longPress_buttonR() {
  Keyboard.press(KEY_RIGHT_CTRL);
} // End of longPress_buttonR

void longPressStop_buttonR() {
  Keyboard.releaseAll();
} // end of longPressStop_buttonR

// BUTTON L FUNCTIONS
void singleClick_buttonL() {
  //Place button L single click action here
} // End of singleClick_buttonL

void doubleClick_buttonL() {
  changeColor();
} // End of doubleClick_buttonL

void longPress_buttonL() {
   Keyboard.press(KEY_RIGHT_SHIFT);
} // End of longPress_buttonL

void longPressStop_buttonL() {
  Keyboard.releaseAll();
} // end of longPressStop_buttonL

// END OF BUTTON FUNCTIONS

// NEOPIXEL COLOR FUNCTION

void blue(){
  boardPixels->setPixelColor(0, boardPixels->Color(0,0, 24));
  boardPixels->setPixelColor(1, boardPixels->Color(0, 0, 24));
  boardPixels->setPixelColor(2, boardPixels->Color(0, 0, 24));
  boardPixels->setPixelColor(3, boardPixels->Color(0, 0, 24));
  boardPixels->setPixelColor(4, boardPixels->Color(0, 0, 24));
  ringPixels->setPixelColor(0, ringPixels->Color(0, 0, 24));
  ringPixels->setPixelColor(1, ringPixels->Color(0, 0, 24));
  ringPixels->setPixelColor(2, ringPixels->Color(0, 0, 24));
  boardPixels->show();   // Send the updated pixel colors to the hardware.
  ringPixels->show();    // Send the updated pixel colors to the hardware.
}

void orange(){
  boardPixels->setPixelColor(0, boardPixels->Color(20,4, 0));
  boardPixels->setPixelColor(1, boardPixels->Color(20,4, 0));
  boardPixels->setPixelColor(2, boardPixels->Color(20,4, 0));
  boardPixels->setPixelColor(3, boardPixels->Color(20,4, 0));
  boardPixels->setPixelColor(4, boardPixels->Color(20,4, 0));
  ringPixels->setPixelColor(0, ringPixels->Color(20,4, 0));
  ringPixels->setPixelColor(1, ringPixels->Color(20,4, 0));
  ringPixels->setPixelColor(2, ringPixels->Color(20,4, 0));
  boardPixels->show();   // Send the updated pixel colors to the hardware.
  ringPixels->show();   // Send the updated pixel colors to the hardware.
}

void white(){
    boardPixels->setPixelColor(0, boardPixels->Color(8,8, 8));
    boardPixels->setPixelColor(1, boardPixels->Color(8,8, 8));
    boardPixels->setPixelColor(2, boardPixels->Color(8,8, 8));
    boardPixels->setPixelColor(3, boardPixels->Color(8,8, 8));
    boardPixels->setPixelColor(4, boardPixels->Color(8,8, 8));
    ringPixels->setPixelColor(0, ringPixels->Color(8,8, 8));
    ringPixels->setPixelColor(1, ringPixels->Color(8,8, 8));
    ringPixels->setPixelColor(2, ringPixels->Color(8,8, 8));
    boardPixels->show();   // Send the updated pixel colors to the hardware.
    ringPixels->show();   // Send the updated pixel colors to the hardware.
}

void off(){
  boardPixels->clear(); // Set all keyboard pixel colors to 'off'
  ringPixels->clear();
  boardPixels->show();   // Send the updated pixel colors to the hardware}.
  ringPixels->show();  // Send the updated pixel colors to the hardware.
}

// CHANGE PIXEL COLORS
void pixelColor(){
  boardPixels->clear(); // Set all keyboard pixel colors to 'off'
  ringPixels->clear(); // Set all ring pixel colors to 'off'

  if (color == 0){ // BLUE
    blue();
  }

  else if (color == 1){ // ORANGE
     orange();
  }

  else if (color == 2){ // WHITE
    white();
  }

  else if (color == 3){ // OFF
    off();
  }

}

// COLOR CHOICE BUTTON COUNTER
void changeColor(){
  if (color == 0){
    color = 1;
  }

  else if (color == 1){
    color = 2;
  }

  else if (color == 2){
    color = 3;
  }

  else if (color == 3){
    color = 0;
  }
}

// STARTUP COLOR ANIMATION
void startUpColor(){
  //boardPixels->clear(); // Set all keyboard pixel colors to 'off'
  //ringPixels->clear(); // Set all ring pixel colors to 'off'
  blue();
  delay(300); 
  orange();
  delay(300);
  white();
  delay(300);
  off();
  delay(50);
}

// ############ END OF FUNCTIONS ###############
